#include <ncbi.h>
#include <blast.h>

static BLAST_DBFilePtr LIBCALL blast_db_open_pro_fasta PROTO((BLAST_DBFilePtr, CharPtr name, BLAST_Alphatype));
static BLAST_Error LIBCALL blast_db_getseq_pro_fasta PROTO((BLAST_DBFilePtr, BLAST_SeqPtr));
static BLAST_StrPtr LIBCALL blast_db_curname_pro_fasta PROTO((BLAST_DBFilePtr));
static void LIBCALL blast_db_close_pro_fasta PROTO((BLAST_DBFilePtr));

static BLAST_DBFilePtr LIBCALL blast_db_open_pro_ig PROTO((BLAST_DBFilePtr, CharPtr name, BLAST_Alphatype));
static BLAST_Error LIBCALL blast_db_getseq_pro_ig PROTO((BLAST_DBFilePtr, BLAST_SeqPtr));
static BLAST_StrPtr LIBCALL blast_db_curname_pro_ig PROTO((BLAST_DBFilePtr));
static void LIBCALL blast_db_close_pro_ig PROTO((BLAST_DBFilePtr));


static BLAST_DBProcSet dbproclist[] = {
		{
			NULL,
			blast_db_open_pro_fasta,
			blast_db_close_pro_fasta,
			blast_db_getseq_pro_fasta,
			blast_db_curname_pro_fasta,
			NULL,
			NULL,
			NULL,
			NULL,
			blast_alphatype_amino_acid,
			BLAST_DB_FMT_FASTA,
			0, 0
		}
	};


BLAST_Error LIBCALL
BlastDBProcSetRegister(dbpsp)
	BLAST_DBProcSetPtr	dbpsp;
{
	register BLAST_DBProcSetPtr	dp = &dbproclist[0];

	while (dp->next != NULL) {
		if (dp->atype == dbpsp->atype &&
				dp->format == dbpsp->format &&
				dp->version_min == dbpsp->version_min &&
				dp->version_max == dbpsp->version_max) {
			/* replace existing entry */
			dbpsp->next = dp->next;
			Nlm_MemCpy((VoidPtr)dp, (VoidPtr)dbpsp, sizeof(*dp));
			return BLAST_ERR_NONE;
		}
		dp = dp->next;
	}
	/* append new entry */
	dbpsp = (BLAST_DBProcSetPtr)Nlm_MemDup(dbpsp, sizeof(*dbpsp));
	dp->next = dbpsp;
	dbpsp->next = NULL;
	return BLAST_ERR_NONE;
}

BLAST_DBProcSetPtr LIBCALL
BlastDBProcSetFind(atype, format, version)
	BLAST_Alphatype	atype;
	BLAST_DBFormat	format;
	long	version;
{
	register BLAST_DBProcSetPtr	dp;

	for (dp = &dbproclist[0]; dp != NULL; dp = dp->next) {
		if ((atype == 0 || dp->atype == atype) &&
			(format == 0 || dp->format == format) &&
			(version == 0 ||
					(dp->version_min <= version &&
						dp->version_max >= version)) )
				return dp;
	}
	return NULL;
}

/*
	BlastDBProcSetFindNext

	Return a pointer to the next BLAST_DBProcSet following dp in the list
	that matches the specified criteria.
*/
BLAST_DBProcSetPtr LIBCALL
BlastDBProcSetFindNext(dp, atype, format, version)
	BLAST_DBProcSetPtr	dp;
	BLAST_Alphatype	atype;
	BLAST_DBFormat	format;
	long	version;
{
	int		i;

	if (dp == NULL)
		dp = &dbproclist[0];
	for (; dp != NULL; dp = dp->next) {
		if ((atype == 0 || dp->atype == atype) &&
			(format == 0 || dp->format == format) &&
			(version == 0 ||
				(dp->version_min <= version && dp->version_max >= version)) )
				return dp;
	}
	return NULL;
}

BLAST_DBFilePtr LIBCALL
BlastDBOpen(name, atype, format, mode)
	CharPtr	name;
	BLAST_Alphatype	atype;
	BLAST_DBFormat	format;
	CharPtr	mode;
{
	BLAST_DBFilePtr	dbfp;
	BLAST_DBProcSetPtr	dbpsp = NULL;

	for (dbpsp = BlastDBProcSetFind(atype, format, 0); dbpsp != NULL;
				dbpsp = BlastDBProcSetFindNext(dbpsp, atype, format, 0) ) {
		if ((dbfp = (dbpsp->open)(NULL, name, atype, format, mode)) != NULL)
			return dbfp;
	}
	return NULL;
}

unsigned long LIBCALL
BlastDBRecordID(dbfp)
	BLAST_DBFilePtr	dbfp;
{
	return dbfp->recid;
}


BLAST_Error LIBCALL
BlastDBGetSeq(dbfp, sp)
	BLAST_DBFilePtr	dbfp;
	BLAST_StrPtr	sp;
{
	return (dbfp->dbpsp->getseq)(dbfp, sp);
}

BLAST_StrPtr LIBCALL
BlastDBCurname(dbfp)
	BLAST_DBFilePtr	dbfp;
{
	return (dbfp->dbpsp->curname)(dbfp);
}


void LIBCALL
BlastDBClose(dbfp)
	BLAST_DBFilePtr	dbfp;
{
	(dbfp->dbpsp->close)(dbfp);
	return;
}


typedef struct fasta_blk {
	MFILE	*mfp;
	BLAST_AlphaMapPtr	amp_in, amp_out;
	UcharPtr	seq;
	size_t	seqlen, seqmax;
	CharPtr	name;
	size_t	namelen, namemax;
	CharPtr	iobuf;
	long	offset0; /* offset of start of first record in file */
	} FastaBlk, PNTR FastaBlkPtr;

static BLAST_DBFilePtr LIBCALL
blast_db_open_pro_fasta(dbfp, name, ap, format)
	BLAST_DBFilePtr	dbfp;
	CharPtr	name;
	BLAST_AlphabetPtr	ap;
	BLAST_DBFormat	format;
{
	BLAST_DBProcSetPtr	psp;
	BLAST_AlphabetPtr	asciip;
	FastaBlkPtr	fbp = NULL;
	FILE	*fp = NULL;
	int		ch;

	if (dbfp != NULL || ap != BLAST_ALPHATYPE_AMINO_ACID)
		return NULL;

	dbfp = (BLAST_DBFilePtr)BlastCalloc(sizeof(*dbfp));
	if (dbfp == NULL)
		return NULL;
	psp = blast_db_procset(atype, format, 0);
	if (psp == NULL) {
		blast_db_close_pro_fasta(dbfp);
		return NULL;
	}
	*dbfp->dbpsp = *psp;

	fbp = (FastaBlkPtr)BlastCalloc(sizeof(*fbp));
	dbfp->procsetdata = fbp;
	if (fbp == NULL)
		goto ErrReturn;

	dbfp->ap = ap;
	asciip = BlastAlphabetFindByName("PRINT");
	fbp->amp_in = BlastAlphaMapFind(asciip, ap);
	fbp->amp_out = BlastAlphaMapFind(ap, asciip);
	fbp->mfp = fp = mfil_open(name, mode, MFIL_OPT_ALL);
	if (fp == NULL)
		goto ErrReturn;

#ifndef WIN_MSWIN
	{
		size_t	bbuflen;
		CharPtr	iobuf;

		iobuflen = 32*KBYTE;
		fbp->iobuf = (CharPtr)BlastMalloc(iobuflen);
		if (fbp->iobuf != NULL) {
			if (setvbuf(fp, fbp->iobuf, _IOFBF, iobuflen) != 0) {
				BlastFree(fbp->iobuf);
				fbp->iobuf = NULL;
			}
		}
	}
#endif

	while ((ch = getc(fp)) != EOF && ch != '>') {
		while ((ch = getc(fp)) != EOF)
			if (ch == '\n')
				break;
		if (ch == EOF)
			break;
		if ((ch = getc(fp)) == EOF)
			break;
	}
	if (ch != '>') {
		blast_db_close_pro_fasta(dbfp);
		return NULL;
	}
	ungetc('>', fp);
	fbp->offset0 = ftell(fp);

	return dbfp; /* normal return */

ErrReturn:
	if (blast_errno == BLAST_ERR_NONE)
		blast_errno = BLAST_ERR_UNDEF;
	blast_db_close_pro_fasta(dbfp);
	return NULL;
}

static void LIBCALL
blast_db_close_pro_fasta(dbfp)
	BLAST_DBFilePtr	dbfp;
{
	FastaBlkPtr	fbp;

	if (dbfp == NULL)
		return;
	fbp = (FastaBlkPtr)dbfp->procsetdata;
	if (fbp != NULL) {
		if (fbp->fp != NULL)
			(void) mfil_close(fbp->fp);
		if (fbp->iobuf != NULL)
			BlastFree(fbp->iobuf);
		Nlm_MemSet(fbp, 0, sizeof(*fbp));
		BlastFree(fbp);
	}
	Nlm_MemSet(dbfp, 0, sizeof(*dbfp));
	BlastFree(dbfp);
	return;
}

static BLAST_Error LIBCALL
blast_db_getseq_pro_fasta(dbfp, sp)
	BLAST_DBFilePtr	dbfp;
	BLAST_StrPtr	sp;
{
	FastaBlkPtr	fbp;
	MFILE	*fp;
	size_t	len;

	if (dbfp->error != 0)
		goto ErrReturn;

	fbp = dbfp->procsetdata;
	fp = fbp->fp;

	dbfp->recid = sp->id.i = (unsigned long)ftell(fp);

	if (get_fasta(fbp->amp_in, &fbp->name, &fbp->namemax, &fbp->namelen, sp, fp) != 0)
		goto ErrReturn;

	dbfp->nrec++;
	dbfp->totlen += sp->len;

	return BLAST_ERR_NONE;

ErrReturn:
	dbfp->error = 1;
	return BLAST_ERR_DBASE_BIGERR;
}

static BLAST_StrPtr	LIBCALL
blast_db_curname_pro_fasta(dbfp)
	BLAST_DBFilePtr	dbfp;
{
	return &dbfp->header;
}


get_fasta(amp, name, namemax, namelen, sp, mfp)
	BLAST_AlphaMapPtr	amp;
	char	**name;
	size_t	*namemax;
	size_t	*namelen;
	BLAST_StrPtr	sp;
	MFILE	*mfp;
{
	CharPtr	buf0;
	size_t	*bufmax;
	BLAST_AlphabetPtr	from;
	CharPtr	buf;
	register Nlm_Boolean	PNTR maptst = amp->maptst;
	register BLAST_LetterPtr	map = amp->map;
	register CharPtr	bufp, cp;
	register int	ch;
	size_t	len;

	if (MFILE_SP(mfp) == NULL)
		return fget_fasta(amp, name, namemax, namelen, sp, mfp->fp);
	while ((ch = getc(fp)) != '>') {
		switch (ch) {
		case '\n': case '\r': case ' ': case '\t':
			continue;
		case EOF:
			return 1;
		default:
			fprintf(stderr, "\nFile is not in FASTA format\n");
			exit(6);
		}
	}
	if (vfgets(name, namemax, name, fp) == NULL)
		return 1;
	len = strlen(*name);
	if (len > 0 && (*name)[len-1] == '\n')
		--len;
	(*name)[*namelen = len] = '\0';

	if (sp->str == NULL) {
		sp->_str = (BLAST_LetterPtr)Nlm_Malloc(2*sizeof(sp->_str[0]));
		if (sp->_str == NULL)
			return 1;
		sp->str = sp->_str + 1;
		sp->alloclen = 2*sizeof(sp->_str[0]);
		sp->len = sp->elen = 0;
	}
	sp->ap = amp->to;

	sp->_str[0] = sp->ap->sentinel;
	buf0 = (CharPtr)sp->_str;
	cp = buf = buf0 + sizeof(sp->_str[0]);
	while ((ch = getc(fp)) != '>' && ch != EOF) {
		ungetc(ch, fp);
		if (vfgets(&buf0, &sp->alloclen, &buf, fp) == NULL)
			break;
		cp = buf;
		while (maptst[ch = *cp]) {
			*cp++ = (char)map[ch];
			if (maptst[ch = *cp] == 0)
				break;
			*cp++ = (char)map[ch];
		}
		if (ch != '\n' && ch != NULLB) {
			bufp = cp;
			for (;;) {
				while (maptst[ch = *++bufp])
					*cp++ = (char)map[ch];
				if (ch == '\n' || ch == NULLB)
					break;
			}
		}
		buf = cp;
	}
	if (ch == '>')
		ungetc(ch, fp);
	sp->_str = (BLAST_LetterPtr)buf0;
	sp->str = sp->_str + 1;
#if BLAST_LETTER_SIZE == 1
	*cp = amp->to->sentinel;
	sp->maxlen = sp->alloclen;
	sp->len = sp->elen = cp - buf0 - 1;
#else
	sp->len = sp->elen = cp - buf0 - sizeof(BLAST_Letter);
	if (sp->len > sp->alloclen / sizeof(BLAST_Letter)) {
		size_t	len = (sp->len + 2) * sizeof(BLAST_Letter);

		sp->_str = Nlm_Realloc(sp->_str, len);
		if (sp->_str == NULL)
			return 1;
		sp->alloclen = len;
	}
	sp->maxlen = sp->alloclen / sizeof(BLAST_Letter);
	{
		register BLAST_LetterPtr	lp, lpmin;

		lpmin = sp->str;
		lp = lpmin + sp->len;
		*lp = amp->to->sentinel;
		cp = ((CharPtr)sp->str) + sp->len;
		while (lp > lpmin)
			*--lp = (BLAST_Letter)*--cp;
	}
#endif
	sp->lpb = 1;
	return sp->len == 0;
}
}

int
fget_fasta(amp, name, namemax, namelen, sp, fp)
	BLAST_AlphaMapPtr	amp;
	char	**name;
	size_t	*namemax;
	size_t	*namelen;
	BLAST_StrPtr	sp;
	FILE	*fp;
{
	CharPtr	buf0;
	size_t	*bufmax;
	BLAST_AlphabetPtr	from;
	CharPtr	buf;
	register Nlm_Boolean	PNTR maptst = amp->maptst;
	register BLAST_LetterPtr	map = amp->map;
	register CharPtr	bufp, cp;
	register int	ch;
	size_t	len;

	while ((ch = getc(fp)) != '>') {
		switch (ch) {
		case '\n': case '\r': case ' ': case '\t':
			continue;
		case EOF:
			return 1;
		default:
			fprintf(stderr, "\nFile is not in FASTA format\n");
			exit(6);
		}
	}
	if (vfgets(name, namemax, name, fp) == NULL)
		return 1;
	len = strlen(*name);
	if (len > 0 && (*name)[len-1] == '\n')
		--len;
	(*name)[*namelen = len] = '\0';

	if (sp->str == NULL) {
		sp->_str = (BLAST_LetterPtr)Nlm_Malloc(2*sizeof(sp->_str[0]));
		if (sp->_str == NULL)
			return 1;
		sp->str = sp->_str + 1;
		sp->alloclen = 2*sizeof(sp->_str[0]);
		sp->len = sp->elen = 0;
	}
	sp->ap = amp->to;

	sp->_str[0] = sp->ap->sentinel;
	buf0 = (CharPtr)sp->_str;
	cp = buf = buf0 + sizeof(sp->_str[0]);
	while ((ch = getc(fp)) != '>' && ch != EOF) {
		ungetc(ch, fp);
		if (vfgets(&buf0, &sp->alloclen, &buf, fp) == NULL)
			break;
		cp = buf;
		while (maptst[ch = *cp]) {
			*cp++ = (char)map[ch];
			if (maptst[ch = *cp] == 0)
				break;
			*cp++ = (char)map[ch];
		}
		if (ch != '\n' && ch != NULLB) {
			bufp = cp;
			for (;;) {
				while (maptst[ch = *++bufp])
					*cp++ = (char)map[ch];
				if (ch == '\n' || ch == NULLB)
					break;
			}
		}
		buf = cp;
	}
	if (ch == '>')
		ungetc(ch, fp);
	sp->_str = (BLAST_LetterPtr)buf0;
	sp->str = sp->_str + 1;
#if BLAST_LETTER_SIZE == 1
	*cp = amp->to->sentinel;
	sp->maxlen = sp->alloclen;
	sp->len = sp->elen = cp - buf0 - 1;
#else
	sp->len = sp->elen = cp - buf0 - sizeof(BLAST_Letter);
	if (sp->len > sp->alloclen / sizeof(BLAST_Letter)) {
		size_t	len = (sp->len + 2) * sizeof(BLAST_Letter);

		sp->_str = Nlm_Realloc(sp->_str, len);
		if (sp->_str == NULL)
			return 1;
		sp->alloclen = len;
	}
	sp->maxlen = sp->alloclen / sizeof(BLAST_Letter);
	{
		register BLAST_LetterPtr	lp, lpmin;

		lpmin = sp->str;
		lp = lpmin + sp->len;
		*lp = amp->to->sentinel;
		cp = ((CharPtr)sp->str) + sp->len;
		while (lp > lpmin)
			*--lp = (BLAST_Letter)*--cp;
	}
#endif
	sp->lpb = 1;
	return sp->len == 0;
}

int
put_fasta(amp, prefix, prefixlen, name, namelen, sp, fp)
	BLAST_AlphaMapPtr	amp;
	CharPtr	prefix;
	size_t	prefixlen;
	CharPtr	name;
	size_t	namelen;
	BLAST_StrPtr	sp;
	FILE	*fp;
{
	BLAST_LetterPtr	seq;
	size_t	seqlen;
	size_t	len;
	int		i;
	char	buf[60];

	if (amp->from != sp->ap || sp->lpb != 1)
		return 1;
	if (putc('>', fp) != '>')
		return 1;
	if (prefix != NULL) {
		fwrite(prefix, prefixlen, 1, fp);
		putc(':', fp);
	}
	fwrite(name, namelen, 1, fp);
	putc('\n', fp);
	seq = sp->str;
	seqlen = sp->len;
	while (seqlen > 0) {
		len = MIN(sizeof(buf), seqlen);
		for (i = 0; i < len; ++i)
			buf[i] = amp->map[seq[i]];
			/*buf[i] = (char)BlastAlphaMapChr(amp, seq[i]);*/
		fwrite(buf, len, 1, fp);
		putc('\n', fp);
		seq += len;
		seqlen -= len;
	}
	return 0;
}
