#include <ncbi.h>
#include <blast/blast.h>
/*
  BlastCheckSpan()
   discards HSPs that are spanned.  An HSP is said to "contain" another,
if one (or both) of the segments in the segment pair spans the entirety of the
corresponding segment in the other HSP.

The span algorithm ensures that two HSPs will never appear together in the
output if one of them "contains" the other.  An old HSP may be replaced by a
new HSP if either HSP contains the other AND:  (a) the new HSP has a higher
score (or higher information contribution); or (b) the new HSP has the same
score but is shorter.  Conversely, the new HSP will be discarded if either HSP
contains the other AND:  (a) the old HSP has a higher score; or (b) the old HSP
has the same score but is shorter.

In any case, a newly found HSP is never used to discard an old HSP until
it is known that the new HSP should not itself be discarded due to some
other old HSP.
*/

int LIBCALL
BlastCheckSpan(hlp, span, hp)
	BLAST_HitListPtr	hlp;
	int		(*span)();
	BLAST_HSPPtr	hp;
{
	BLAST_HSPPtr	hp2, firstspan;
	register BLAST_HSPPtr	hpnext, hpprev;

	/* Make sure the new HSP is not excluded by an HSP already in the list */
	for (hp2 = *hlp->hpp, firstspan = NULL; hp2 != NULL; hp2 = hp2->next) {
		if ((*span)(hp, hp2) == TRUE) {
			if (BlastHSPCmpByNormScore(&hp, &hp2) < 0) {
				if (firstspan == NULL)
					firstspan = hp2;
				continue;
			}
			/* There is already a better HSP in the list */
			hp->wfcp->stats.spanned++;
			return 1;
		}
	}

	if (firstspan == NULL) {
		/* No spanning at all -- save the new HSP and return */
		hp->wfcp->stats.saved++;
		if (BlastHSPSave(hlp, hp) != NULL)
			return 0;
		return 1;
	}

	firstspan->wfcp->stats.spanned++;
	firstspan->wfcp->stats.saved--;
	hp->wfcp->stats.saved++;
	/* Save the new HSP in existing storage, overwriting the spanned HSP */
	hp->next = firstspan->next; /* prepare to copy back this important value */
	*firstspan = *hp;
	firstspan->hlp = hlp;

	/* Discard any other HSPs that are spanned by the new one */
	for (hpprev = firstspan, hp2 = firstspan->next; hp2 != NULL; hp2 = hpnext) {
		hpnext = hp2->next;
		if ((*span)(hp, hp2) == TRUE) {
			hp2->wfcp->stats.spanned++;
			hp2->wfcp->stats.saved--;
			hpprev->next = hpnext;
			BlastHSPPut(hp2);
			continue;
		}
		hpprev = hp2;
	}
	return 0;
}



/*-------------------------  span0()---------------------------------

	span0(hp1, hp2)

	Always returns FALSE (indicating no spanning).
	Use this function to prevent rejection of spanned HSPs.
*/
int LIBCALL
span0(hp1, hp2)
	BLAST_HSPPtr	hp1, hp2;
{
	return FALSE;
}

/*-------------------------  span1()---------------------------------

	span1(hp1, hp2)

	return TRUE if hp1 and hp2 span on either sequence (or both);
	otherwise return FALSE.
*/
#ifndef CODON_LEN
#define CODON_LEN	3
#endif

int LIBCALL
span1(hp1, hp2)
	BLAST_HSPPtr	hp1, hp2;
{
	BLAST_Coord	pos10, pos11, pos20, pos21;
	int		qstrand1, qstrand2, sstrand1, sstrand2, f1, f2;

	/* Return FALSE if hits are on different strands */
	qstrand1 = SIGN(hp1->q_seg.frame);
	qstrand2 = SIGN(hp2->q_seg.frame);
	if (qstrand1 != qstrand2)
		return FALSE;
	sstrand1 = SIGN(hp1->s_seg.frame);
	sstrand2 = SIGN(hp2->s_seg.frame);
	if (sstrand1 != sstrand2)
		return FALSE;

	/* Look first for spanning on subject/database sequence... */

	f1 = hp1->s_seg.frame;
	f2 = hp2->s_seg.frame;
	if (f1 == f2) {
		pos10 = hp1->s_seg.offset;
		pos20 = hp2->s_seg.offset;
		pos11 = pos10 + hp1->len;
		pos21 = pos20 + hp2->len;
	}
	else {
		f1 = ABS(f1);
		f2 = ABS(f2);
		pos10 = hp1->s_seg.offset * CODON_LEN + f1;
		pos20 = hp2->s_seg.offset * CODON_LEN + f2;
		pos11 = pos10 + hp1->len * CODON_LEN;
		pos21 = pos20 + hp2->len * CODON_LEN;
	}

	if (pos10 <= pos20 && pos11 >= pos21)
		return TRUE;
	if (pos20 <= pos10 && pos21 >= pos11)
		return TRUE;


	/* Look for spanning on the query sequence... */

	f1 = hp1->q_seg.frame;
	f2 = hp2->q_seg.frame;
	if (f1 == f2) {
		pos10 = hp1->q_seg.offset;
		pos20 = hp2->q_seg.offset;
		pos11 = pos10 + hp1->len;
		pos21 = pos20 + hp2->len;
	}
	else {
		f1 = ABS(f1);
		f2 = ABS(f2);
		pos10 = hp1->q_seg.offset * CODON_LEN + f1;
		pos20 = hp2->q_seg.offset * CODON_LEN + f2;
		pos11 = pos10 + hp1->len * CODON_LEN;
		pos21 = pos20 + hp2->len * CODON_LEN;
	}

	if (pos10 <= pos20 && pos11 >= pos21)
		return TRUE;
	if (pos10 >= pos20 && pos11 <= pos21)
		return TRUE;

	return FALSE;
}

/*-------------------------  span2()---------------------------------

	span2(hp1, hp2)

	return TRUE iff hp1 and hp2 span on both the 1st and 2nd sequences;
	otherwise return FALSE.
*/
#ifndef CODON_LEN
#define CODON_LEN	3
#endif

int LIBCALL
span2(hp1, hp2)
	BLAST_HSPPtr	hp1, hp2;
{
	BLAST_Coord	pos10, pos1f, pos20, pos2f;
	int		qstrand1, qstrand2, sstrand1, sstrand2, f1, f2;

	/* Return FALSE if hits are on different strands */
	qstrand1 = SIGN(hp1->q_seg.frame);
	qstrand2 = SIGN(hp2->q_seg.frame);
	if (qstrand1 != qstrand2)
		return FALSE;
	sstrand1 = SIGN(hp1->s_seg.frame);
	sstrand2 = SIGN(hp2->s_seg.frame);
	if (sstrand1 != sstrand2)
		return FALSE;

	/* Look first for spanning on subject/database sequence... */

	f1 = hp1->s_seg.frame;
	f2 = hp2->s_seg.frame;
	if (f1 == f2) {
		pos10 = hp1->s_seg.offset;
		pos20 = hp2->s_seg.offset;
		pos1f = pos10 + hp1->len;
		pos2f = pos20 + hp2->len;
	}
	else {
		f1 = ABS(f1);
		f2 = ABS(f2);
		pos10 = hp1->s_seg.offset * CODON_LEN + f1;
		pos20 = hp2->s_seg.offset * CODON_LEN + f2;
		pos1f = pos10 + hp1->len * CODON_LEN;
		pos2f = pos20 + hp2->len * CODON_LEN;
	}

	if (pos10 < pos20 && pos1f < pos2f)
		return FALSE;
	if (pos10 > pos20 && pos1f > pos2f)
		return FALSE;


	/* Look for spanning on the query sequence... */

	f1 = hp1->q_seg.frame;
	f2 = hp2->q_seg.frame;
	if (f1 == f2) {
		pos10 = hp1->q_seg.offset;
		pos20 = hp2->q_seg.offset;
		pos1f = pos10 + hp1->len;
		pos2f = pos20 + hp2->len;
	}
	else {
		f1 = ABS(f1);
		f2 = ABS(f2);
		pos10 = hp1->q_seg.offset * CODON_LEN + f1;
		pos20 = hp2->q_seg.offset * CODON_LEN + f2;
		pos1f = pos10 + hp1->len * CODON_LEN;
		pos2f = pos20 + hp2->len * CODON_LEN;
	}

	if (pos10 < pos20 && pos1f < pos2f)
		return FALSE;
	if (pos10 > pos20 && pos1f > pos2f)
		return FALSE;

	return TRUE;
}
